// Library for manipulating line via user interface
// efg, August 1998
// www.efg2.com/lab2

unit gULineLibrary;

interface

uses
  Windows, // TPoint, ClipCursor
  Forms, // TForm
  ExtCtrls; // TImage

type
  TLineSelected = (lsNotSelected, lsPoint1, lsPoint2, lsLine);

    // TLineOrientation is used as part of the heuristic algorithm that decides
    // if a line is selected
  TLineOrientation = (loPoint, loHorizontal, loVertical);

function AddPoints(const PointA, PointB: TPoint): TPoint;
function SubtractPoints(const PointA, PointB: TPoint): TPoint;

procedure CalcLineParameters(const PointA, PointB: TPoint;
  var Slope, Intercept: DOUBLE;
  var LineOrientation: TLineOrientation);
function NearLine(const Target, Point1, Point2: TPoint): BOOLEAN;


procedure RestrictCursorToDrawingArea(const Image: TImage);
procedure RemoveCursorRestrictions;


function SquareContainsPoint(const SquareCenter: TPoint;
  const SquareHalfSize: INTEGER; {pixels}
  const TestPoint: TPoint): BOOLEAN;

implementation

uses
  Math, // MinIntValue, MaxIntValue
  Classes; // Point


function AddPoints(const PointA, PointB: TPoint): TPoint;
begin
  with RESULT do
  begin
    X := PointA.X + PointB.X;
    Y := PointA.Y + PointB.Y
  end
end {AddPoints};


function SubtractPoints(const PointA, PointB: TPoint): TPoint;
begin
  with RESULT do
  begin
    X := PointA.X - PointB.X;
    Y := PointA.Y - PointB.Y
  end
end {SubtractPoints};


  // Determine whether a line is ltHorizonal or ltVertical,  along with the
  // appropriate slope and intercept FOR point-slope line  equations.  These
  // parameters are used to determine if a line is selected.

procedure CalcLineParameters(const PointA, PointB: TPoint;
  var Slope, Intercept: DOUBLE;
  var LineOrientation: TLineOrientation);
var
  Delta: TPoint;
begin
  Delta := SubtractPoints(PointB, PointA);

  if (Delta.X = 0) and (Delta.Y = 0)
    then begin
      // This special CASE should never happen if iMinPixels > 0
    LineOrientation := loPoint;
    Slope := 0.0;
    Intercept := 0.0
  end
  else begin

    if ABS(Delta.X) >= ABS(Delta.Y)
      then begin
        // The line is more horizontal than vertical.  Determine values FOR
        // equation:  Y = slope*X + intercept
      LineOrientation := loHorizontal;
      try
        Slope := Delta.Y / Delta.X {conventional slope in geometry}
      except
        Slope := 0.0
      end;
      Intercept := PointA.Y - PointA.X * Slope
    end
    else begin
        // The line is more vertical than horizontal.  Determine values FOR
        // equation:  X = slope*Y + intercept
      LineOrientation := loVertical;
      try
        Slope := Delta.X / Delta.Y {reciprocal of conventional slope}
      except
        Slope := 0.0
      end;
      Intercept := PointA.X - PointA.Y * Slope;
    end

  end
end {CalcLineParameters};


  // Determine if Target1 is "near" line segment between Point1 and Point2

function NearLine(const Target, Point1, Point2: TPoint): BOOLEAN;
const
  LineSelectFuzz = 4; // Pixel "fuzz" used in line selection
var
  Intercept: DOUBLE;
  LineOrientation: TLineOrientation;
  maxX: INTEGER;
  maxY: INTEGER;
  minX: INTEGER;
  minY: INTEGER;
  Slope: DOUBLE;
  xCalculated: INTEGER;
  yCalculated: INTEGER;
begin
  RESULT := FALSE;

    // If an Endpoint is not selected, was part of line selected?
  CalcLineParameters(Point1, Point2, Slope, Intercept,
    LineOrientation);

  case LineOrientation of
    loHorizontal:
      begin
        minX := MinIntValue([Point1.X, Point2.X]);
        maxX := MaxIntValue([Point1.X, Point2.X]);
          // first check if selection within horizontal range of line
        if (Target.X >= minX) and (Target.X <= maxX)
          then begin
            // Since X is within range of line, now see if Y value is close
            // enough to the calculated Y value FOR the line to be selected.
          yCalculated := ROUND(Slope * Target.X + Intercept);
          if ABS(yCalculated - Target.Y) <= LineSelectFuzz
            then RESULT := TRUE
        end
      end;

    loVertical:
      begin
        minY := MinIntValue([Point1.Y, Point2.Y]);
        maxY := MaxIntValue([Point1.Y, Point2.Y]);
          // first check if selection within vertical range of line
        if (Target.Y >= minY) and (Target.Y <= maxY)
          then begin
            // Since Y is within range of line, now see if X value is close
            // enough to the calculated X value FOR the line to be selected.
          xCalculated := ROUND(Slope * Target.Y + Intercept);
          if ABS(xCalculated - Target.X) <= LineSelectFuzz
            then RESULT := TRUE
        end
      end;

    loPoint:
        // Do nothing -- should not occur
  end
end {NearLine};


  ///////////////////////////////////////////////////////////////////////

procedure RestrictCursorToDrawingArea(const Image: TImage);
var
  CursorClipArea: TRect;
begin
  CursorClipArea := Bounds(Image.ClientOrigin.X,
    Image.ClientOrigin.Y,
    Image.Width, Image.Height);
  Windows.ClipCursor(@CursorClipArea)
end {RestrictCursorToDrawingArea};


procedure RemoveCursorRestrictions;
begin
  Windows.ClipCursor(nil)
end {RemoveCursorRestrictions};


  ///////////////////////////////////////////////////////////////////////

  // Could use Windows PtInRect API call instead (but it excludes the
  // bottom and right edges of the rectangle while including the left
  // and top edges).

function SquareContainsPoint(const SquareCenter: TPoint;
  const SquareHalfSize: INTEGER; {pixels}
  const TestPoint: TPoint): BOOLEAN;
begin
  RESULT := (TestPoint.X >= SquareCenter.X - SquareHalfSize) and
    (TestPoint.X <= SquareCenter.X + SquareHalfSize) and
    (TestPoint.Y >= SquareCenter.Y - SquareHalfSize) and
    (TestPoint.Y <= SquareCenter.Y + SquareHalfSize)
end {SquareContainsPoint};


end.
