unit szScanThread;

interface

uses
   Windows, SysUtils, Classes, Forms, ComCtrls;

type
   TDirInfo = class;

   TBaseInfo = class
      Name        : string;
      Size        : double;
      Wasted      : double;      // Total bytes wasted in this dir and subdirs.  "Wasted" space = unused space in the last cluster
      Parent      : TDirInfo;
      function    FullPath: string;
   end;

   TFileInfo = class(TBaseInfo)
      LastWrite   : TDateTime;
      CheckSum    : integer;     // Adler-32 Checksum - for duplicate processing
      HasChkSum   : boolean;     // The checksum was calculated for this file
      ChkMatch    : boolean;     // Checksum matches all of its duplicates' checksum
   end;

   TDirInfo = class(TBaseInfo)
      TreeNode    : TTreeNode;
      NumDirs     : integer;
      NumFiles    : integer;
      SubDirs     : TList;
      Files       : TList;
      constructor Create;
      destructor  Destroy; override;
      // Total bytes in this directory and all sub-directories (does not include Wasted space)
      property    TotBytes: double read Size write Size;
   end;

   TScanThread = class(TThread)
   protected
      StatMsg     : string;
      NumDir      : integer;
      procedure   Execute; override;
      function    ProcessDir(const Base, Dir: string): TDirInfo;
      procedure   SendStatus;
   public
      MinClus     : integer;
      MainForm    : TForm;
      Drive       : string;
      Directory   : string;
      VolName     : string;
      Root        : TDirInfo;
      AllFiles    : TList;
      ErrMsg      : string;
      destructor  Destroy; override;
   end;

implementation

uses szMain;


{ TBaseInfo }

function TBaseInfo.FullPath: string;
var
   tmp : TBaseInfo;
begin
   Result := Name;

   tmp := Parent;
   while (tmp<>nil) do begin
      Result := tmp.Name + '\' + Result;
      tmp := tmp.Parent;
   end;
end;


{ TDirInfo }

constructor TDirInfo.Create;
begin
   inherited;
   SubDirs := TList.Create;
   Files := TList.Create;
end;

destructor TDirInfo.Destroy;
var
   i : integer;
begin
   for i := 0 to SubDirs.Count-1 do TObject(SubDirs[i]).Free;
   for i := 0 to Files.Count-1 do TObject(Files[i]).Free;
   SubDirs.Free;
   Files.Free;
   inherited;
end;


{ TScanThread }

procedure TScanThread.SendStatus;
begin
   TszMainForm(szMainForm).Stat[2] := StatMsg;
end;


function TScanThread.ProcessDir(const Base, Dir: string): TDirInfo;
var
   sr        : TSearchRec;
   FileInfo  : TFileInfo;
   SubDir    : TDirInfo;
   tmp, tmp2 : integer;
   ft        : TFileTime;
   st        : TSystemTime;
begin
   Result := TDirInfo.Create;

   if Dir='' then Result.Name := Base
      else Result.Name := Dir;

   if FindFirst(Base + '\*.*', faAnyFile, sr) = 0 then repeat
      if Terminated then begin
         FindClose(sr);
         exit;
      end;

      if (sr.Name = '.') or (sr.Name = '..') then continue;

      if (sr.Attr and faDirectory)<>0 then begin
         // ** DIRECTORY **
         SubDir            := ProcessDir(Base + '\' + sr.Name, sr.Name);
         SubDir.Parent     := Result;
         Result.TotBytes   := Result.TotBytes   + SubDir.TotBytes;
         Result.Wasted     := Result.Wasted + SubDir.Wasted;
         Result.NumDirs    := Result.NumDirs + 1 + SubDir.NumDirs;
         Result.NumFiles   := Result.NumFiles + SubDir.NumFiles;
         Result.SubDirs.Add(SubDir);
      end else begin
         // ** FILE **
         FileInfo := TFileInfo.Create;
         FileInfo.Name := sr.Name;
         tmp := sr.Size;
         tmp2 := tmp mod MinClus;
         if tmp2 > 0 then
            tmp2 := MinClus - tmp2;
         FileInfo.Size := tmp;
         try
            FileTimeToLocalFileTime(sr.FindData.ftLastWriteTime, ft);
            FileTimeToSystemTime(ft, st);
            if st.wYear < 2300 then
               FileInfo.LastWrite := SystemTimeToDateTime(st);
         except
            FileInfo.LastWrite := 0;
         end;
         FileInfo.Parent := Result;
         FileInfo.Wasted := tmp2;
         Result.Files.Add(FileInfo);
         AllFiles.Add(FileInfo);

         Result.TotBytes := Result.TotBytes + tmp + tmp2;
         Result.Wasted   := Result.Wasted + tmp2;
         Result.NumFiles := Result.NumFiles + 1;
      end;
   until FindNext(sr)<>0;
   FindClose(sr);

   NumDir := NumDir + 1;
   StatMsg := Format('%1.0n: %s', [NumDir+0.0, Base]);
   Synchronize(SendStatus);
end;

procedure TScanThread.Execute;
var
   Path     : string;
   VolBuf   : array[0..MAX_PATH] of char;
   SecPerClus, BytePerSec, FreeClus, TotClus : DWord;{integer;}
   MaxLen, Flags  : DWord;{integer;}
begin
   AllFiles := TList.Create;
   try
      AllFiles.Capacity := 32768;
      Path := Drive + ':';
{function GetDiskFreeSpace(lpRootPathName: PChar;
  var lpSectorsPerCluster, lpBytesPerSector,
     lpNumberOfFreeClusters, lpTotalNumberOfClusters: DWORD): BOOL; stdcall;}

      GetDiskFreeSpace(PChar(Path+'\'), SecPerClus, BytePerSec, FreeClus, TotClus);
      MinClus := SecPerClus * BytePerSec;
{function GetVolumeInformation(lpRootPathName: PChar;
  lpVolumeNameBuffer: PChar; nVolumeNameSize: DWORD; lpVolumeSerialNumber: PDWORD;
  var lpMaximumComponentLength, lpFileSystemFlags: DWORD;
  lpFileSystemNameBuffer: PChar; nFileSystemNameSize: DWORD): BOOL; stdcall;}

      Win32Check(GetVolumeInformation(PChar(Path+'\'), VolBuf, sizeof(VolBuf), nil, MaxLen, Flags, nil, 0));
      VolName := VolBuf;
      VolName := VolName + ' (' + Path + ')';

      NumDir := 0;
      Root := ProcessDir(Directory, '');
   except
      on e: Exception do ErrMsg := e.Message;
   end;
end;

destructor TScanThread.Destroy;
begin
   // no need to free it, the main thread takes ownership
   // Root.Free;
   // AllFiles.Free;
end;


end.
